/*
<:copyright-broadcom 
 
 Copyright (c) 2004 Broadcom Corporation 
 All Rights Reserved 
 No portions of this material may be reproduced in any form without the 
 written permission of: 
          Broadcom Corporation 
          16215 Alton Parkway 
          Irvine, California 92619 
 All information contained in this document is Broadcom Corporation 
 company private, proprietary, and trade secret. 
 
:>
*/
//**************************************************************************
// File Name  : bcmmii.c
//
// Description: Broadcom PHY/Ethernet Switch Configuration
//               
//**************************************************************************

#include <linux/types.h>
#include <linux/delay.h>
#include <linux/mii.h>
#include <linux/stddef.h>

#include "bcmenet.h"

#if defined(CONFIG_BCM96338)
#include <6338_map.h>
#endif
#if defined(CONFIG_BCM96348)
#include <6348_map.h>
#endif
#if defined(CONFIG_BCM96345)
#include <6345_map.h>
#endif
#include "bcmenet.h"
#include "bcmmii.h"

typedef enum {
    MII_100MBIT     = 0x0001,
    MII_FULLDUPLEX  = 0x0002,
    MII_AUTONEG     = 0x0004,
}   MII_CONFIG;

#define EMAC_MDC            0x1f

/* local prototypes */
static MII_CONFIG mii_getconfig(struct net_device *dev);
static MII_CONFIG mii_autoconfigure(struct net_device *dev);
static void mii_setup(struct net_device *dev);
static void mii_soft_reset(struct net_device *dev, int PhyAddr);
#if defined (CONFIG_BCM96345)
static void gpioouten(gpio_reg_addrs_t *sbh, uint32 mask, uint32 val);
static void gpioout(gpio_reg_addrs_t *sbh, uint32 mask, uint32 val);
static uint32 gpioin(gpio_reg_addrs_t *sbh);
static void gpio_attach(ethsw_info_t *ethsw);
static void gpio_write(ethsw_info_t *ethsw, uint8 *buf, int len);
static void gpio_read(ethsw_info_t *ethsw, uint8 *buf, int len);
static void ethsw_gpio_select(struct net_device *dev, uint8 page);
static void ethsw_gpio_wreg(struct net_device *dev, uint8 page, uint8 addr, uint8 *buf, int len);
static void ethsw_gpio_rreg(struct net_device *dev, uint8 page, uint8 addr, uint8 *buf, int len);
#endif
#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
static int spi_read( int spi_ss, unsigned char *msg_buf, int nbytes );
static int spi_write(int spi_ss, unsigned char *msg_buf, int nbytes );
static int ethsw_spi_ss_id(struct net_device *dev);
static void ethsw_spi_select(struct net_device *dev, int page);
static void ethsw_spi_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
static void ethsw_spi_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
#endif
static void ethsw_mdio_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
static void ethsw_mdio_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
static void ethsw_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
static void ethsw_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len);
static int ethsw_switch_type(struct net_device *dev);
#if defined (CONFIG_BCM96338)
static void pga_fix_enable(struct net_device *dev);
#endif
#ifdef PHY_LOOPBACK
static void mii_loopback(struct net_device *dev);
#endif


/* read a value from the MII */
int mii_read(struct net_device *dev, int phy_id, int location) 
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac = pDevCtrl->emac;

    emac->mdioData = MDIO_RD | (phy_id << MDIO_PMD_SHIFT) | (location << MDIO_REG_SHIFT);
    mdelay(1);
    while ( ! (emac->intStatus & EMAC_MDIO_INT) );
    emac->intStatus |= EMAC_MDIO_INT;
    return emac->mdioData & 0xffff;
}

/* write a value to the MII */
void mii_write(struct net_device *dev, int phy_id, int location, int val)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac = pDevCtrl->emac;
    uint32 data;

    data = MDIO_WR | (phy_id << MDIO_PMD_SHIFT) | (location << MDIO_REG_SHIFT) | val;
    emac->mdioData = data;
    mdelay(1);
    while ( ! (emac->intStatus & EMAC_MDIO_INT) );
    emac->intStatus |= EMAC_MDIO_INT;
}

#ifdef PHY_LOOPBACK
/* set the MII loopback mode */
static void mii_loopback(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    uint32 val;

    TRACE(("mii_loopback\n"));

    val = mii_read(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_BMCR);
    /* Disable autonegotiation */
    val &= ~BMCR_ANENABLE;
    /* Enable Loopback */
    val |= BMCR_LOOPBACK;
    mii_write(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_BMCR, val);
}
#endif

int mii_linkstatus(struct net_device *dev, int phy_id)
{
    int val;

    val = mii_read(dev, phy_id, MII_BMSR);
    /* reread: the link status bit is sticky */
    val = mii_read(dev, phy_id, MII_BMSR);

    if (val & BMSR_LSTATUS)
        return 1;
    else
        return 0;
}

void mii_linkstatus_start(struct net_device *dev, int phy_id)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac = pDevCtrl->emac;
    int location = MII_BMSR;
    
    emac->mdioData = MDIO_RD | (phy_id << MDIO_PMD_SHIFT) |
        (location << MDIO_REG_SHIFT);
}

int mii_linkstatus_check(struct net_device *dev, int *up)
{
    int done = 0;
    int val;
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac = pDevCtrl->emac;

    if( emac->intStatus & EMAC_MDIO_INT ) {

        emac->intStatus |= EMAC_MDIO_INT;
        val = emac->mdioData & 0xffff;
        *up = ((val & BMSR_LSTATUS) == BMSR_LSTATUS) ? 1 : 0;
        done = 1;
    }

    return( done );
}

void mii_enablephyinterrupt(struct net_device *dev, int phy_id)
{
    mii_write(dev, phy_id, MII_INTERRUPT, 
        MII_INTR_ENABLE | MII_INTR_MASK_FDX | MII_INTR_MASK_LINK_SPEED);
}

void mii_clearphyinterrupt(struct net_device *dev, int phy_id)
{
    mii_read(dev, phy_id, MII_INTERRUPT);
}

/* return the current MII configuration */
static MII_CONFIG mii_getconfig(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    uint32 val;
    MII_CONFIG eConfig = 0;

    TRACE(("mii_getconfig\n"));

    /* Read the Link Partner Ability */
    val = mii_read(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_LPA);
    if (val & LPA_100FULL) {          /* 100 MB Full-Duplex */
        eConfig = (MII_100MBIT | MII_FULLDUPLEX);
    } else if (val & LPA_100HALF) {   /* 100 MB Half-Duplex */
        eConfig = MII_100MBIT;
    } else if (val & LPA_10FULL) {    /* 10 MB Full-Duplex */
        eConfig = MII_FULLDUPLEX;
    } 

    return eConfig;
}

/* Auto-Configure this MII interface */
static MII_CONFIG mii_autoconfigure(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    int i;
    int val;
    MII_CONFIG eConfig;

    TRACE(("mii_autoconfigure\n"));

    /* enable and restart autonegotiation */
    val = mii_read(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_BMCR);
    val |= (BMCR_ANRESTART | BMCR_ANENABLE);
    mii_write(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_BMCR, val);

    /* wait for it to finish */
    for (i = 0; i < 1000; i++) {
        mdelay(1);
        val = mii_read(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_BMSR);
        if (val & BMSR_ANEGCOMPLETE) {
            break;
        }
    }

    eConfig = mii_getconfig(dev);
    if (val & BMSR_ANEGCOMPLETE) {
        eConfig |= MII_AUTONEG;
    } 

    return eConfig;
}

static void mii_setup(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    MII_CONFIG eMiiConfig;

    eMiiConfig = mii_autoconfigure(dev);

    if (! (eMiiConfig & MII_AUTONEG)) {
        printk(KERN_INFO CARDNAME ": Auto-negotiation timed-out\n");
    }

    if ((eMiiConfig & (MII_100MBIT | MII_FULLDUPLEX)) == (MII_100MBIT | MII_FULLDUPLEX)) {
        printk(KERN_INFO CARDNAME ": 100 MB Full-Duplex (auto-neg)\n");
    } else if (eMiiConfig & MII_100MBIT) {
        printk(KERN_INFO CARDNAME ": 100 MB Half-Duplex (auto-neg)\n");
    } else if (eMiiConfig & MII_FULLDUPLEX) {
        printk(KERN_INFO CARDNAME ": 10 MB Full-Duplex (auto-neg)\n");
    } else {
        printk(KERN_INFO CARDNAME ": 10 MB Half-Duplex (assumed)\n");
    }

#ifdef PHY_LOOPBACK
    /* Enable PHY loopback */
    mii_loopback(dev);
#endif

    /* Check for FULL/HALF duplex */
    if (eMiiConfig & MII_FULLDUPLEX) {
        pDevCtrl->emac->txControl = EMAC_FD;
    }
}

/* reset the MII */
static void mii_soft_reset(struct net_device *dev, int PhyAddr) 
{
    int val;

    mii_write(dev, PhyAddr, MII_BMCR, BMCR_RESET);
    udelay(10); /* wait ~10usec */
    do {
        val = mii_read(dev, PhyAddr, MII_BMCR);
    } while (val & BMCR_RESET);

}

#if defined (CONFIG_BCM96338)
static void pga_fix_enable(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    int val;
    int phyaddr;

    phyaddr = pDevCtrl->EnetInfo.ucPhyAddress;
    val = (MII_BRCM_TEST_HARDRESET |
           MII_BRCM_TEST_10BT_SERIAL_NODRIB |
           MII_BRCM_TEST_100TX_POWERDOWN |
           MII_BRCM_TEST_10BT_POWERDOWN );
    mii_write(dev, phyaddr, MII_BRCM_TEST, val); //  reset phy
    val = mii_read(dev, phyaddr, MII_BRCM_TEST);
    val |= MII_BRCM_TEST_SHADOW_ENABLE;
    mii_write(dev, phyaddr, MII_BRCM_TEST, val); //  shadow mode
    val = mii_read(dev, phyaddr, MII_TX_CONTROL);
    val |= MII_TX_CONTROL_PGA_FIX_ENABLE;
    mii_write(dev, phyaddr, MII_TX_CONTROL, val); // pga fix enable
    val = mii_read(dev, phyaddr, MII_BRCM_TEST);
    val &= ~MII_BRCM_TEST_SHADOW_ENABLE;
    mii_write(dev, phyaddr, MII_BRCM_TEST, val); //  shadow mode
}

/* phy_tuning_mse */
void phy_tuning_mse(struct net_device *dev) 
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac = pDevCtrl->emac;
    int phyaddr;
    int val;
    int i = 0;
    int retry = 0;

    phyaddr = pDevCtrl->EnetInfo.ucPhyAddress;
    val = mii_read(dev, phyaddr, MII_BRCM_TEST);
    val |= MII_BRCM_TEST_SHADOW_ENABLE;
    mii_write(dev, phyaddr, MII_BRCM_TEST, val); // enabled shadow mode

    do {
        i++;
        val = mii_read(dev, phyaddr, MII_AUX_STATUS3); // read the MSE value
        if ((val & MII_AUX_STATUS3_MSE_MASK) >= 0x4000) {
            val = mii_read(dev, phyaddr, MII_TX_CONTROL);
            val &= ~MII_TX_CONTROL_PGA_FIX_ENABLE;
            mii_write(dev, phyaddr, MII_TX_CONTROL, val); // pga fix disable
            udelay(100); /* wait ~100usec */
            val |= MII_TX_CONTROL_PGA_FIX_ENABLE;
            mii_write(dev, phyaddr, MII_TX_CONTROL, val); // pga fix enable
            i = 0;
            retry++;
        }
        if ((i > 12) || (retry > 2)) // read twelve times to ensure good 
            break;
    } while (1);

    val = mii_read(dev, phyaddr, MII_BRCM_TEST);
    val &= ~MII_BRCM_TEST_SHADOW_ENABLE;
    mii_write(dev, phyaddr, MII_BRCM_TEST, val); // disabled shadow mode
}
#endif

#if defined (CONFIG_BCM96345)
/* BCM5325 register SPI access through gpio */
/* Set GPIO direction. */
static void gpioouten(gpio_reg_addrs_t *sbh, uint32 mask, uint32 val)
{
    *sbh->gpio_direction_reg = (*sbh->gpio_direction_reg & ~mask) | val;
}

/* Set GPIO value. */
static void gpioout(gpio_reg_addrs_t *sbh, uint32 mask, uint32 val)
{
    *sbh->gpio_value_reg = (*sbh->gpio_value_reg & ~mask) | val;
}

/* Read GPIO value. */
static uint32 gpioin(gpio_reg_addrs_t *sbh)
{
    return( (uint32) *sbh->gpio_value_reg );
}

/* Get access to the Ethernet Switch */
static void gpio_attach(ethsw_info_t *ethsw)
{
    /* Initialize GPIO outputs */
    gpioouten(&ethsw->sbh, ethsw->ssl | ethsw->clk | ethsw->mosi | ethsw->miso,
        ethsw->ssl | ethsw->clk | ethsw->mosi);
}

/* Write a byte stream to the chip */
static void gpio_write(ethsw_info_t *ethsw, uint8 *buf, int len)
{
    int i;
    uint8 mask;

    for (i = 0; i < len; i++) {
        /* Bit bang from MSB to LSB */
        for (mask = 0x80; mask; mask >>= 1) {
            /* Clock low */
            gpioout(&ethsw->sbh, ethsw->clk, 0);
            udelay(10);

            /* Output on rising edge */
            if (mask & buf[i])
                gpioout(&ethsw->sbh, ethsw->mosi, ethsw->mosi);
            else
                gpioout(&ethsw->sbh, ethsw->mosi, 0);

            /* Clock high */
            gpioout(&ethsw->sbh, ethsw->clk, ethsw->clk);
            udelay(10);
        }
    }
}

/* Handy macros for writing fixed length values */
#define gpio_write8(ethsw, b) { uint8 val = (uint8) (b); gpio_write((ethsw), &val, sizeof(val)); }
#define gpio_write16(ethsw, w) { uint16 val = (uint16) (w); gpio_write((ethsw), &val, sizeof(val)); }
#define gpio_write32(ethsw, l) { uint32 val = (uint32) (l); gpio_write((ethsw), &val, sizeof(val)); }

/* Read a byte stream from the chip */
static void gpio_read(ethsw_info_t *ethsw, uint8 *buf, int len)
{
    int i, timeout;
    uint8 rack, mask, byte;

    /* Timeout after 100 tries without RACK */
    for (i = 0, rack = 0, timeout = 100; i < len && timeout;) {
        /* Bit bang from MSB to LSB */
        for (mask = 0x80, byte = 0; mask; mask >>= 1) {
            /* Clock low */
            gpioout(&ethsw->sbh, ethsw->clk, 0);
            udelay(10);

            /* Sample on rising edge */
            if (gpioin(&ethsw->sbh) & ethsw->miso)
                byte |= mask;

            /* Clock high */
            gpioout(&ethsw->sbh, ethsw->clk, ethsw->clk);
            udelay(10);
        }

        /* RACK when bit 0 is high */
        if (!rack) {
            rack = (byte & 1);
            timeout--;
        } else {
            buf[i] = byte;
            i++;
        }
    }
}       

/* Handy macros for reading fixed length values */
#define gpio_read8(ethsw) { uint8 val; gpio_read((ethsw), &val, sizeof(val)); val; }
#define gpio_read16(ethsw) { uint16 val; gpio_read((ethsw), &val, sizeof(val)); val; }
#define gpio_read32(ethsw) { uint32 val; gpio_read((ethsw), &val, sizeof(val)); val; }

static void ethsw_gpio_select(struct net_device *dev, uint8 page)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    ethsw_info_t *ethsw = &pDevCtrl->ethSwitch;

    if (ethsw->cid == pDevCtrl->EnetInfo.ucPhyAddress && ethsw->page == page)
        return;

    ethsw->cid = pDevCtrl->EnetInfo.ucPhyAddress;
    ethsw->page = page;

    /* Enable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, 0);
    udelay(10);

    /* Select new chip */
    gpio_write8(ethsw, BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_WRITE |
        ((ethsw->cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT));

    /* Select new page */
    gpio_write8(ethsw, PAGE_SELECT);
    gpio_write8(ethsw, ethsw->page);

    /* Disable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, ethsw->ssl);
    udelay(10);
}

/* Write chip register */
static void ethsw_gpio_wreg(struct net_device *dev, uint8 page, uint8 addr, uint8 *buf, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    ethsw_info_t *ethsw = &pDevCtrl->ethSwitch;

    /* Select chip and page */
    ethsw_gpio_select(dev, page);

   /* Enable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, 0);
    udelay(10);

    /* Write */
    gpio_write8(ethsw, BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_WRITE |
        ((ethsw->cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT));
    gpio_write8(ethsw, addr);
    gpio_write(ethsw, buf, len);

    /* Disable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, ethsw->ssl);
    udelay(10);
}

/* Read chip register */
static void ethsw_gpio_rreg(struct net_device *dev, uint8 page, uint8 addr, uint8 *buf, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    ethsw_info_t *ethsw = &pDevCtrl->ethSwitch;

    /* Select chip and page */
    ethsw_gpio_select(dev, page);

    /* Enable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, 0);
    udelay(10);

    /* Fast read */
    gpio_write8(ethsw, BCM5325_SPI_CMD_FAST | BCM5325_SPI_CMD_READ |
        ((pDevCtrl->ethSwitch.cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT));
    gpio_write8(ethsw, addr);
    gpio_read(ethsw, buf, len);

    /* Disable CS */
    gpioout(&ethsw->sbh, ethsw->ssl, ethsw->ssl);
    udelay(10);
}
#endif

#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
/* BCM5325 register access through SPI */
static int spi_read( int spi_ss, unsigned char *msg_buf, int nbytes )
{
    int i;

    SPI->spiMsgCtl = (nbytes << SPI_BYTE_CNT_SHIFT |
                            HALF_DUPLEX_R << SPI_MSG_TYPE_SHIFT);
    
    for (i = 0; i < BCM5325_SPI_PREPENDCNT; i++)
	    SPI->spiMsgData[i] = msg_buf[i];

    SPI->spiCmd = (SPI_CMD_START_IMMEDIATE << SPI_CMD_COMMAND_SHIFT | 
                    spi_ss << SPI_CMD_DEVICE_ID_SHIFT | 
                    BCM5325_SPI_PREPENDCNT << SPI_CMD_PREPEND_BYTE_CNT_SHIFT );

    while (!(SPI->spiIntStatus & SPI_INTR_CMD_DONE));

    SPI->spiIntStatus = SPI_INTR_CLEAR_ALL;

    for(i = 0; i < nbytes; i++) {
	    msg_buf[i] = SPI->spiRxDataFifo[i];
    }

    return SPI_STATUS_OK;
}

static int spi_write(int spi_ss, unsigned char *msg_buf, int nbytes )
{
    int i;
    
    if (nbytes < BCM5325_SPI_PREPENDCNT) {
        return SPI_STATUS_INVALID_LEN;
    }

    SPI->spiMsgCtl = (nbytes << SPI_BYTE_CNT_SHIFT |
                    HALF_DUPLEX_W << SPI_MSG_TYPE_SHIFT);

    for (i = 0; i < nbytes; i++)
	    SPI->spiMsgData[i] = msg_buf[i];

    SPI->spiCmd = (SPI_CMD_START_IMMEDIATE << SPI_CMD_COMMAND_SHIFT | 
                    spi_ss << SPI_CMD_DEVICE_ID_SHIFT | 
                    0 << SPI_CMD_PREPEND_BYTE_CNT_SHIFT );

    while (!(SPI->spiIntStatus & SPI_INTR_CMD_DONE));

    SPI->spiIntStatus = SPI_INTR_CLEAR_ALL;

    return SPI_STATUS_OK;
}

static int ethsw_spi_ss_id(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    int slave_select;

    switch(pDevCtrl->EnetInfo.usConfigType) {
        case BP_ENET_CONFIG_SPI_SSB_0:
            slave_select = SPI_DEV_ID_0;
            break;
        case BP_ENET_CONFIG_SPI_SSB_1:
            slave_select = SPI_DEV_ID_1;
            break;
        case BP_ENET_CONFIG_SPI_SSB_2:
            slave_select = SPI_DEV_ID_2;
            break;
        case BP_ENET_CONFIG_SPI_SSB_3:
            slave_select = SPI_DEV_ID_3;
            break;
        default:
            slave_select = SPI_DEV_ID_1;
            break;
    }
    return slave_select;
}

static void ethsw_spi_select(struct net_device *dev, int page)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    ethsw_info_t *ethsw = &pDevCtrl->ethSwitch;
    unsigned char buf[3];
    int spi_ss;

    if (ethsw->cid == pDevCtrl->EnetInfo.ucPhyAddress && ethsw->page == page)
        return;

    spi_ss = ethsw_spi_ss_id(dev);
    ethsw->cid = pDevCtrl->EnetInfo.ucPhyAddress;
    ethsw->page = page;
    /* Select new chip */
    buf[0] = BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_WRITE |
        ((ethsw->cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT);

    /* Select new page */
    buf[1] = PAGE_SELECT;
    buf[2] = (char)ethsw->page;
    spi_write(spi_ss, buf, sizeof(buf));
}

static void ethsw_spi_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    unsigned char buf[64];
    int rc;
    int i;
    int max_check_spi_sts;
    int spi_ss;

    ethsw_spi_select(dev, page);

    spi_ss = ethsw_spi_ss_id(dev);
    /* write command byte and register address */
    buf[0] = BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_READ |
        ((pDevCtrl->ethSwitch.cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT);
    buf[1] = (unsigned char)reg;
    rc = spi_read(spi_ss, buf, 1);

    if (rc == SPI_STATUS_OK) {
        max_check_spi_sts = 0;
        do {
            /* write command byte and read spi_sts address */
            buf[0] = BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_READ |
                ((pDevCtrl->ethSwitch.cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT);
            buf[1] = (unsigned char)BCM5325_SPI_STS;
            rc = spi_read(spi_ss, buf, 1);
            if (rc == SPI_STATUS_OK) {
                /* check the bit 0 RACK bit is set */
                if (buf[0] & BCM5325_SPI_CMD_RACK) {
                    break;
                }
                mdelay(1);
            } else {
                break;
            }
        } while (max_check_spi_sts++ < 10);

        if (rc == SPI_STATUS_OK) {
            for (i = len-1; i >= 0; i--) {
                buf[0] = BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_READ |
                    ((pDevCtrl->ethSwitch.cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT);
                buf[1] = (unsigned char)0xf0;
                rc = spi_read(spi_ss, buf, 1);
                if (rc == SPI_STATUS_OK) {
                    /* Write the data out in LE format to the switch */
                    *(data + len - i - 1) = buf[0];
                }
            }
        }
   }
}

static void ethsw_spi_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    unsigned char buf[64];
    int i;
    int spi_ss;

    ethsw_spi_select(dev, page);

    spi_ss = ethsw_spi_ss_id(dev);
    buf[0] = BCM5325_SPI_CMD_NORMAL | BCM5325_SPI_CMD_WRITE |
        ((pDevCtrl->ethSwitch.cid & BCM5325_SPI_CHIPID_MASK) << BCM5325_SPI_CHIPID_SHIFT);

    buf[1] = (char)reg;
    for (i = 0; i < len; i++) {
        /* Write the data out in LE format to the switch */
        buf[BCM5325_SPI_PREPENDCNT+i] = *(data + (len - i - 1));
    }
    spi_write( spi_ss, buf, len+BCM5325_SPI_PREPENDCNT);
}
#endif

/* BCM5325E register access through MDC/MDIO */
static void ethsw_mdio_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac; 
    int cmd, res, ret;
    int max_retry = 0;

    emac = pDevCtrl->emac;
    cmd = (page << REG_PPM_REG16_SWITCH_PAGE_NUMBER_SHIFT) | REG_PPM_REG16_MDIO_ENABLE;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG16, cmd);

    cmd = (reg << REG_PPM_REG17_REG_NUMBER_SHIFT) | REG_PPM_REG17_OP_READ;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG17, cmd);

    do {
        res = mii_read(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG17);
        udelay(10);
    } while ((max_retry++ < 5) &&
             ((res & (REG_PPM_REG17_OP_WRITE|REG_PPM_REG17_OP_READ)) != REG_PPM_REG17_OP_DONE));

    ret = 0;
    ret |= mii_read(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG24) << 0;
    ret |= mii_read(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG25) << 16;
    switch (len) {
        case 1:
            *data = (uint8)ret;
            break;
        case 2:
            *(uint16 *)data = (uint16)ret;
            break;
        case 4:
            *(uint32 *)data = ret;
            break;
    }
}

static void ethsw_mdio_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac; 
    uint32 cmd, res;
    int val = 0;
    int max_retry = 0;

    emac = pDevCtrl->emac;
    switch (len) {
        case 1:
            val = *data;
            break;
        case 2:
            val = *(uint16 *)data;
            break;
        case 4:
            val = *(uint32 *)data;
            break;
    }
    cmd = (page << REG_PPM_REG16_SWITCH_PAGE_NUMBER_SHIFT) | REG_PPM_REG16_MDIO_ENABLE;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG16, cmd);

    cmd = val>>0 & 0xffff;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG24, cmd);
    cmd = val>>16 & 0xffff;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG25, cmd);
    cmd = 0;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG26, cmd);
    cmd = 0;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG27, cmd);

    cmd = (reg << REG_PPM_REG17_REG_NUMBER_SHIFT) | REG_PPM_REG17_OP_WRITE;
    mii_write(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG17, cmd);

    do {
        res = mii_read(dev, PSEUDO_PHY_ADDR, REG_PSEUDO_PHY_MII_REG17);
        udelay(10);
    } while ((max_retry++ < 5) &&
             ((res & (REG_PPM_REG17_OP_WRITE|REG_PPM_REG17_OP_READ)) != REG_PPM_REG17_OP_DONE));
}

static void ethsw_rreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    switch (pDevCtrl->EnetInfo.usConfigType) {
        case BP_ENET_CONFIG_GPIO:
#if defined (CONFIG_BCM96345)
            ethsw_gpio_rreg(dev, page, reg, data, len);
#endif
            break;

        case BP_ENET_CONFIG_MDIO_PSEUDO_PHY:
            ethsw_mdio_rreg(dev, page, reg, data, len);
            break;

        case BP_ENET_CONFIG_SPI_SSB_0:
        case BP_ENET_CONFIG_SPI_SSB_1:
        case BP_ENET_CONFIG_SPI_SSB_2:
        case BP_ENET_CONFIG_SPI_SSB_3:
#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
            ethsw_spi_rreg(dev, page, reg, data, len);
#endif
            break;
    }
}

static void ethsw_wreg(struct net_device *dev, int page, int reg, uint8 *data, int len)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);

    switch (pDevCtrl->EnetInfo.usConfigType) {
        case BP_ENET_CONFIG_GPIO:
#if defined (CONFIG_BCM96345)
            ethsw_gpio_wreg(dev, page, reg, data, len);
#endif
            break;

        case BP_ENET_CONFIG_MDIO_PSEUDO_PHY:
            ethsw_mdio_wreg(dev, page, reg, data, len);
            break;

        case BP_ENET_CONFIG_SPI_SSB_0:
        case BP_ENET_CONFIG_SPI_SSB_1:
        case BP_ENET_CONFIG_SPI_SSB_2:
        case BP_ENET_CONFIG_SPI_SSB_3:
#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
            ethsw_spi_wreg(dev, page, reg, data, len);
#endif
            break;
    }
}

/* Configure the switch for VLAN support */
static int ethsw_switch_type(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    uint8 val8   = 0;
    int type;

    if ( pDevCtrl->EnetInfo.ucPhyType != BP_ENET_EXTERNAL_SWITCH ||
        ((pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_GPIO) &&
         (pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_MDIO_PSEUDO_PHY) &&
         (pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_SPI_SSB_0) &&
         (pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_SPI_SSB_1) &&
         (pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_SPI_SSB_2) &&
         (pDevCtrl->EnetInfo.usConfigType != BP_ENET_CONFIG_SPI_SSB_3)) )
         return ESW_TYPE_UNDEFINED;

    /*
     * Determine if the external PHY is connected to a 5325E, 5325F or 5325M
     * read CTRL REG 4 if the default value is 
     * 1 - 5325E
     * 3 - 5325F
     * else - 5325M
     */ 
    ethsw_rreg(dev, PAGE_VLAN, REG_VLAN_CTRL4, (char*)&val8, sizeof(val8));
    switch (val8)
    {
    case 1: /* BCM5325E */
        type = ESW_TYPE_BCM5325E;
        break;
    case 3: /* BCM5325F */
        type = ESW_TYPE_BCM5325F;
        break;
    default: /* default to BCM5325M */
        type = ESW_TYPE_BCM5325M;
    }
    return type;
}

/* Configure the switch for VLAN support */
int ethsw_config_vlan(struct net_device *dev, int enable, unsigned int vid)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    uint8 val8   = 0;
    uint16 val16 = 0;
    uint32 val32 = 0;
    uint port;
    int max_retry = 0;

    switch(pDevCtrl->ethSwitch.type) {
        case ESW_TYPE_BCM5325E:
        case ESW_TYPE_BCM5325F:
            break;
        default:
            return -EFAULT;
    }

    /* Enable 802.1Q VLAN and Individual VLAN learning mode */
    if (enable) {
        val8 = REG_VLAN_CTRL0_ENABLE_1Q | REG_VLAN_CTRL0_IVLM;
    }
    else
    {
        val8 = (uint8) ~REG_VLAN_CTRL0_ENABLE_1Q;
    }
    ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_CTRL0, (char *)&val8, sizeof(val8));

    /* Enable high order 8 bit VLAN table check and drop non .1Q frames for MII port */
    val8 = REG_VLAN_CTRL3_8BIT_CHECK | REG_VLAN_CTRL3_MII_DROP_NON_1Q;
    ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_CTRL3, (char *)&val8, sizeof(val8));

    /* Drop V_table miss frames */
    val8 = REG_VLAN_CTRL5_DROP_VTAB_MISS;
    ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_CTRL5, (char *)&val8, sizeof(val8));

    for (port = 0; port < 5; port++) {
        /* default tag for every port */
        val16 = vid + port;
        ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_PTAG0 + port * sizeof(uint16), (char *)&val16, sizeof(val16));

        /* Program VLAN table for each VLAN */
        val32 = ((1 << MII_PORT) | (1 << port)) << REG_VLAN_GROUP_SHIFT;
        val32 |= (1 << port) << REG_VLAN_UNTAG_SHIFT;
        val32 |= REG_VLAN_WRITE_VALID;
        ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_WRITE, (char*)&val32, sizeof(val32));
        val16 = REG_VLAN_ACCESS_START_DONE | REG_VLAN_ACCESS_WRITE_STATE | (vid + port);
        ethsw_wreg(dev, PAGE_VLAN, REG_VLAN_ACCESS, (char*)&val16, sizeof(val16));
        do {
            ethsw_rreg(dev, PAGE_VLAN, REG_VLAN_ACCESS, (char*)&val16, sizeof(val16));
            udelay(10);
        } while ((max_retry++ < 5) && (val16 & REG_VLAN_ACCESS_START_DONE));
    }

    return 0;
}

int mii_init(struct net_device *dev)
{
    BcmEnet_devctrl *pDevCtrl = netdev_priv(dev);
    volatile EmacRegisters *emac;
    int data32;
    uint16 data16;
#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
    uint16 blkEnables;
#endif
    uint8 data8;
    int i;
    char *phytype = "";
    char *setup = "";

    switch(pDevCtrl->EnetInfo.ucPhyType) {
        case BP_ENET_INTERNAL_PHY:
            phytype = "Internal PHY";
            break;

        case BP_ENET_EXTERNAL_PHY:
            phytype = "External PHY";
            break;

        case BP_ENET_EXTERNAL_SWITCH:
            phytype = "Ethernet Switch";
            break;

        default:
            printk(KERN_INFO CARDNAME ": Unknown PHY type\n");
            return -1;
    }
    switch (pDevCtrl->EnetInfo.usConfigType) {
        case BP_ENET_CONFIG_GPIO:
            setup = "GPIO";
            break;

        case BP_ENET_CONFIG_MDIO_PSEUDO_PHY:
            setup = "MDIO Sseudo PHY Interface";
            break;

        case BP_ENET_CONFIG_SPI_SSB_0:
            setup = "SPI Slave Select 0";
            break;

        case BP_ENET_CONFIG_SPI_SSB_1:
            setup = "SPI Slave Select 1";
            break;

        case BP_ENET_CONFIG_SPI_SSB_2:
            setup = "SPI Slave Select 2";
            break;

        case BP_ENET_CONFIG_SPI_SSB_3:
            setup = "SPI Slave Select 3";
            break;

        case BP_ENET_CONFIG_MDIO:
            setup = "MDIO";
            break;

        default:
            setup = "Undefined Interface";
            break;
    }
    printk("Config %s Through %s\n", phytype, setup);

    emac = pDevCtrl->emac;
    switch(pDevCtrl->EnetInfo.ucPhyType) {

        case BP_ENET_INTERNAL_PHY:
            /* init mii clock, do soft reset of phy, default is 10Base-T */
            emac->mdioFreq = EMAC_MII_PRE_EN | EMAC_MDC;
            /* reset phy */
            mii_soft_reset(dev, pDevCtrl->EnetInfo.ucPhyAddress);
#if defined(CONFIG_BCM96338)
            pga_fix_enable(dev);
#endif
            mii_setup(dev);
            break;

        case BP_ENET_EXTERNAL_PHY:
            emac->config |= EMAC_EXT_PHY;
            emac->mdioFreq = EMAC_MII_PRE_EN | EMAC_MDC;
            /* reset phy */
            if (pDevCtrl->EnetInfo.usGpioPhyReset != BP_NOT_DEFINED) {
                data16 = GPIO_NUM_TO_MASK(pDevCtrl->EnetInfo.usGpioPhyReset);
                GPIO->GPIODir |= data16;
                if ((pDevCtrl->EnetInfo.usGpioPhyReset & BP_ACTIVE_MASK) == BP_ACTIVE_LOW) {
                    GPIO->GPIOio &= ~data16;
                    udelay(400); /* hold > 150us */
                    GPIO->GPIOio |= data16;
                }
                else {
                    GPIO->GPIOio |= data16;
                    udelay(400); /* hold > 150us */
                    GPIO->GPIOio &= ~data16;
                }
                mdelay(1100); /* wait > 1 second */
            } else {
                mii_soft_reset(dev, pDevCtrl->EnetInfo.ucPhyAddress);
            }

            data32 = mii_read(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_ADVERTISE);
            data32 |= ADVERTISE_FDFC; /* advertise flow control capbility */
            mii_write(dev, pDevCtrl->EnetInfo.ucPhyAddress, MII_ADVERTISE, data32);

            mii_setup(dev);
            break;

        case BP_ENET_EXTERNAL_SWITCH:
            emac->config |= EMAC_EXT_PHY;
            emac->mdioFreq = EMAC_MII_PRE_EN | EMAC_MDC;
            emac->txControl = EMAC_FD;
            switch (pDevCtrl->EnetInfo.usConfigType) {
                case BP_ENET_CONFIG_GPIO:
#if defined (CONFIG_BCM96345)
                    pDevCtrl->ethSwitch.sbh.gpio_direction_reg = (volatile uint16 *)&GPIO->GPIODir;
                    pDevCtrl->ethSwitch.sbh.gpio_value_reg = (volatile uint16 *)&GPIO->GPIOio;
                    pDevCtrl->ethSwitch.ssl = GPIO_NUM_TO_MASK(pDevCtrl->EnetInfo.usGpioPhySpiSs);
                    pDevCtrl->ethSwitch.clk = GPIO_NUM_TO_MASK(pDevCtrl->EnetInfo.usGpioPhySpiSck);
                    pDevCtrl->ethSwitch.mosi = GPIO_NUM_TO_MASK(pDevCtrl->EnetInfo.usGpioPhySpiMosi);
                    pDevCtrl->ethSwitch.miso = GPIO_NUM_TO_MASK(pDevCtrl->EnetInfo.usGpioPhySpiMiso);
                    pDevCtrl->ethSwitch.cid = 0xff;
                    pDevCtrl->ethSwitch.page = 0xff;
                    gpio_attach(&pDevCtrl->ethSwitch);
#endif
                    break;

                case BP_ENET_CONFIG_MDIO_PSEUDO_PHY:
                    mii_soft_reset(dev, PSEUDO_PHY_ADDR);
                    break;

                case BP_ENET_CONFIG_SPI_SSB_0:
                case BP_ENET_CONFIG_SPI_SSB_1:
                case BP_ENET_CONFIG_SPI_SSB_2:
                case BP_ENET_CONFIG_SPI_SSB_3:
#if defined (CONFIG_BCM96338) || defined (CONFIG_BCM96348)
                    blkEnables = PERF->blkEnables;
                    if ((blkEnables & SPI_CLK_EN) == 0) {
                        blkEnables |= SPI_CLK_EN;
                        PERF->blkEnables = blkEnables;
                    }
                    pDevCtrl->ethSwitch.cid = 0xff;
                    pDevCtrl->ethSwitch.page = 0xff;
#endif
                    break;

                case BP_ENET_CONFIG_MDIO:
                    /* reset phy */
                    if (pDevCtrl->EnetInfo.numSwitchPorts) {
                        for (i = 0; i < pDevCtrl->EnetInfo.numSwitchPorts; i++) {
                            mii_soft_reset(dev, pDevCtrl->EnetInfo.ucPhyAddress | i);
                        }
                    }
                    return 0;

                default:
                    printk(KERN_INFO CARDNAME ": Unknown PHY configuration type\n");
                    break;
            }
            if (pDevCtrl->EnetInfo.numSwitchPorts) {
                data8 = 0;
                switch (pDevCtrl->EnetInfo.numSwitchPorts) {
                    case 5:
                        data8 |= REG_POWER_DOWN_MODE_PORT5_PHY_DISABLE;
                    case 4:
                        data8 |= REG_POWER_DOWN_MODE_PORT4_PHY_DISABLE;
                    case 3:
                        data8 |= REG_POWER_DOWN_MODE_PORT3_PHY_DISABLE;
                    case 2:
                        data8 |= REG_POWER_DOWN_MODE_PORT2_PHY_DISABLE;
                    case 1:
                        data8 |= REG_POWER_DOWN_MODE_PORT1_PHY_DISABLE;
                        break;
                    default:
                        break;
                }
                /* disable Switch port PHY */
                ethsw_wreg(dev, PAGE_CONTROL, REG_POWER_DOWN_MODE, (uint8 *)&data8, sizeof(data8));
                /* enable Switch port PHY */
                data8 &= ~data8;
                ethsw_wreg(dev, PAGE_CONTROL, REG_POWER_DOWN_MODE, (uint8 *)&data8, sizeof(data8));
            }

            /* setup Switch MII1 port state override */
            ethsw_rreg(dev, PAGE_CONTROL, REG_CONTROL_MII1_PORT_STATE_OVERRIDE, &data8, sizeof(data8));

            if (pDevCtrl->EnetInfo.usReverseMii == BP_ENET_REVERSE_MII)
                data8 |= REG_CONTROL_MPSO_REVERSE_MII;
            data8 |= (REG_CONTROL_MPSO_MII_SW_OVERRIDE|REG_CONTROL_MPSO_LINKPASS);
            data8 |= (REG_CONTROL_MPSO_LP_FLOW_CONTROL|REG_CONTROL_MPSO_SPEED100|REG_CONTROL_MPSO_FDX);

            ethsw_wreg(dev, PAGE_CONTROL, REG_CONTROL_MII1_PORT_STATE_OVERRIDE, &data8, sizeof(data8));

            /* checking Switch functional */
            data8 = 0;
            ethsw_rreg(dev, PAGE_CONTROL, REG_CONTROL_MII1_PORT_STATE_OVERRIDE, &data8, sizeof(data8));
            if ((data8 & (REG_CONTROL_MPSO_MII_SW_OVERRIDE|REG_CONTROL_MPSO_LINKPASS)) !=
                (REG_CONTROL_MPSO_MII_SW_OVERRIDE|REG_CONTROL_MPSO_LINKPASS) ||
                (data8 == 0xff)) {
                    printk(KERN_INFO CARDNAME ": error on Ethernet Switch setup\n");
                    return -1;
            }
            if (pDevCtrl->EnetInfo.usReverseMii == BP_ENET_REVERSE_MII) {
                if ((data8 & REG_CONTROL_MPSO_REVERSE_MII) != REG_CONTROL_MPSO_REVERSE_MII) {
                    printk(KERN_INFO CARDNAME ": error on Ethernet Switch reverse MII setup\n");
                    return -1;
                }
            }
            pDevCtrl->ethSwitch.type = ethsw_switch_type(dev);

            break;

        default:
            break;
    }

    return 0;
}

